<?php

namespace App\Admin;

use Sonata\AdminBundle\Admin\AbstractAdmin;
use Sonata\AdminBundle\Datagrid\DatagridMapper;
use Sonata\AdminBundle\Datagrid\ListMapper;
use Sonata\AdminBundle\Datagrid\ProxyQueryInterface;
use Sonata\AdminBundle\Form\FormMapper;
use Sonata\AdminBundle\Show\ShowMapper;
use Sonata\AdminBundle\Form\Type\ModelType;
use Sonata\AdminBundle\Form\Type\CollectionType;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use App\Form\PedidoLineasType;
use Symfony\Component\Security\Core\Authorization\AuthorizationCheckerInterface;
use Symfony\Component\Security\Core\Security;

final class PedidosAdmin extends AbstractAdmin
{

    private $authorizationChecker;
    private $security;

    public function __construct(AuthorizationCheckerInterface $authorizationChecker, Security $security)
    {
        $this->authorizationChecker = $authorizationChecker;
        $this->security = $security;
    }

    protected function configureQuery(ProxyQueryInterface $query): ProxyQueryInterface
    {

        if ($this->authorizationChecker->isGranted('ROLE_ADMIN') || $this->authorizationChecker->isGranted('ROLE_SUPER_ADMIN')) {
            $query->orderBy($query->getRootAlias() . '.id', 'DESC');
        } elseif ($this->authorizationChecker->isGranted('ROLE_CLIENTE')) {
            $user = $this->security->getUser();
            $clienteId = $user->getCliente()->getId();
            $query
                ->andWhere($query->expr()->eq($query->getRootAlias() . '.cliente', ':clienteId'))
                ->orderBy($query->getRootAlias() . '.id', 'DESC')
                ->setParameter('clienteId', $clienteId);
        } elseif ($this->authorizationChecker->isGranted('ROLE_FABRICANTE')) {
            $user = $this->security->getUser();
            $fabricanteId = $user->getFabricante()->getId();
            $query
                ->andWhere($query->expr()->eq($query->getRootAlias() . '.fabricante', ':fabricanteId'))
                ->orderBy($query->getRootAlias() . '.id', 'DESC')
                ->setParameter('fabricanteId', $fabricanteId);
        }

        return $query;
    }

    protected function configureDatagridFilters(DatagridMapper $filter): void
    {
        $filter
            ->add('codigoPostal')
            ->add('ciudad')
            ->add('banco')
            ->add('dni')
            ->add('estado')
            ->add('createdAt')
        ;
    }

    protected function configureListFields(ListMapper $list): void
    {
        $list
            ->add('identificador')
            ->add('codigoPostal')
            ->add('ciudad')
            ->add('banco')
            ->add('dni')
            ->add('estado', null, ['template' => 'admin/estados_pedido.html.twig'])
            ->add('createdAt', null, ['format' => 'd/m/Y'])
            ->add('total')
            ->add('_action', 'actions', [
                'actions' => [
                    'show' => [],
                    'edit' => [],
                    'imprimir' => ['template' => 'admin/imprimir_pedido.html.twig'],
                    'descargar' => ['template' => 'admin/descargar_pedido.html.twig'],
                ],
                'label' => 'Acciones',
            ]);
    }

    protected function configureFormFields(FormMapper $form): void
    {
        $form
            ->tab('Datos')
                ->with('Pedido')
                    ->add('estado', ChoiceType::class, [
                        'choices' => [
                            'Pendiente' => 'pendiente',
                            'Validado' => 'validado',
                            'Preparando' => 'preparando',
                            'Enviando' => 'enviando',
                            'Enviado' => 'enviado',
                            'Cancelado' => 'cancelado',
                        ],
                    ])
                    ->add('identificador')
                    ->add('direccion')
                    ->add('codigoPostal')
                    ->add('ciudad')
                    ->add('condiciones')
                    ->add('partes')
                    ->add('banco')
                    ->add('dni')
                    ->add('cuentaBanco')
                    ->add('observaciones')
                    ->add('fabricante', ModelType::class, [
                        'class' => 'App\Entity\Fabricantes',
                        'property' => 'nombre',
                        'required' => true,
                        'btn_add' => false,
                        'label' => 'Fabricante',
                    ])
                    ->add('cliente', ModelType::class, [
                        'class' => 'App\Entity\Clientes',
                        'property' => 'nombre',
                        'required' => false,
                        'btn_add' => false,
                        'label' => 'Cliente',
                    ])
            ->end()
            ->end()
            ->tab('Productos')
                ->with('')
                    ->add('lineas', CollectionType::class, [
                        'entry_type' => PedidoLineasType::class,
                        'allow_add' => true,
                        'by_reference' => false,
                    ])
            ->end()
            ->end()
        ;
    }

    protected function configureShowFields(ShowMapper $show): void
    {
        $show
            ->with('Datos', ['class' => 'col-sm-6'])
                ->add('identificador')
                ->add('direccion')
                ->add('codigoPostal')
                ->add('ciudad')
                ->add('condiciones')
                ->add('partes')
                ->add('banco')
                ->add('cuentaBanco')
                ->add('observaciones')
                ->add('estado')
            ->end()
            ->with('Historial', ['class' => 'col-sm-6'])
                ->add('historial', null, [
                    'template' => 'admin/pedido_historial.html.twig'
                ])
            ->end()
            ->with('Productos', ['class' => 'col-sm-12'])
                ->add('lineas', null, [
                    'template' => 'admin/pedido_lineas.html.twig'
                ])
            ->end()
            ->with('Cliente', ['class' => 'col-sm-6'])
                ->add('cliente.nombre', null, ['label' => 'Nombre'])
                ->add('cliente.telefono', null, ['label' => 'Teléfono'])
                ->add('cliente.email', null, ['label' => 'Email'])
                ->add('dni')
            ->end()
            ->with('Fabricante', ['class' => 'col-sm-6'])
                ->add('fabricante.nombre', null, ['label' => 'Nombre'])
                ->add('fabricante.telefono', null, ['label' => 'Teléfono'])
                ->add('fabricante.email', null, ['label' => 'Email'])
            ->end()
        ;
    }
}